/******************************************************************************
* C Source: Astrometric Calibration                                           *
*                                                                             *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*    This set of functions are used for applying the astrometric of a plate   *
*    to ultimately compute the x-y position on a plate from a given right     *
*    ascension and declination. These functions were either originally        *
*    written, or ported to C by R. White.                                     *
*                                                                             *
* Module List:                                                                *
*    amdinv                                                                   *
*       Computes x-y position on a plate using CALOBCC solution.              *
*    amdpos                                                                   *
*       Routine to convert x,y to RA,Dec using the CALOBCC solution.          *
*    pltmodel                                                                 *
*       Computes plate model and it's partial derivatives.                    *
*    ppoinv                                                                   *
*       Computes x-y position on a plate using PDS orientation solution.      *
*    xypos                                                                    *
*       Computes x-y from right ascension and declination.                    *
*                                                                             *
* History:                                                                    *
*    Created    : 13-AUG-93 by J. Doggett,                                    *
*       for the Association of Universities for Research in Astronomy, Inc.   *
*          at the Space Telescope Science Institute,                          *
*       to facilitate the accessing of data from the CD-ROM set of the        *
*          Digitized Sky Survey.                                              *
*    Delivered  : 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
*                                                                             *
*    Updated    : 01-AUG-94 by J. Doggett,                                    *
*       Updated pltmodel function.                                            *
*    Redelivered: 30-NOV-94 by J. Doggett,                                    *
*       GetImage Version 1.1.                                                 *
*                                                                             *
* Copyright (c) 1993, 1994, Association of Universities for Research in       *
* Astronomy, Inc.  All Rights Reserved.                                       *
* Produced under National Aeronautics and Space Administration grant          *
* NAG W-2166.                                                                 *
******************************************************************************/
#include <math.h>

#include "astrmcal.h"

#include "astronmy.h"
#include "header.h"

/******************************************************************************
* Module:                                                                     *
*    amdinv                                                                   *
*                                                                             *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*    Computes the x,y position on the plate from the right ascension and      *
*    declination using the inverse of the CALOBCC solution. Newton's method   *
*    is used to iterate from a starting position until convergence is         *
*    reached.                                                                 *
*                                                                             *
* Return Value:                                                               *
*    none                                                                     *
*                                                                             *
* Arguments:                                                                  *
*                                                                             *
*    Input Arguments:                                                         *
*       header                                                                *
*          HEADER, Reference                                                  *
*             Structure with header information.                              *
*       ra                                                                    *
*          double, Value                                                      *
*             The right asension of the coordinates to convert to x-y.        *
*       dec                                                                   *
*          double, Value                                                      *
*             The declination of the coordinates to convert to x-y.           *
*       mag                                                                   *
*          float, Value                                                       *
*             Magnitude for the calibration solution.                         *
*       col                                                                   *
*          float, Value                                                       *
*             Color for the calibration solution.                             *
*    Input/Output Arguments:                                                  *
*       none                                                                  *
*    Output Arguments:                                                        *
*       x                                                                     *
*          float, Reference                                                   *
*             Result x position.                                              *
*       y                                                                     *
*          float, Reference                                                   *
*             Result y position.                                              *
*                                                                             *
* History:                                                                    *
*    Created    : 07-MAY-90 by R. White,                                      *
*       Converted from GASP IDL routine to Fortran.                           *
*    Delivered  : 30-JUL-91 by R. White,                                      *
*       Converted from Fortran to C.                                          *
*                                                                             *
*    Updated    : 13-AUG-93 by J. Doggett,                                    *
*       Adopted into software used to facilitate the accessing of data from   *
*          the CD-ROM set of the Digitized Sky Survey.                        *
*    Updated    : 14-DEC-93 by J. Doggett,                                    *
*       Modified to reflect change in calling sequence to traneqstd function. *
*    Redelivered: 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
******************************************************************************/
extern void
amdinv(header,ra,dec,mag,col,x,y)
HEADER header;
double ra, dec;
float mag, col;
float *x, *y;

/****************
* amdinv page 2 *
****************/

{
int i, max_iterations;
float tolerance;
double xi, eta, object_x, object_y, delta_x, delta_y, f, fx, fy, g, gx, gy;

    /*
     *  Initialize
     */
    i = 0;
    max_iterations = 50;
    tolerance = 0.0000005;
    delta_x = tolerance;
    delta_y = tolerance;
    /*
     *  Convert RA and Dec to St.coords
     */
    traneqstd(header->plt_center_ra,header->plt_center_dec,ra,dec,&xi,&eta);
    /*
     *  Set initial value for x,y
     */
    object_x = xi/header->plt_scale;
    object_y = eta/header->plt_scale;
    /*
     *  Iterate by Newtons method
     */
    for(i = 0; i < max_iterations; i++) {
        pltmodel(header,object_x,object_y,mag,col,&f,&fx,&fy,&g,&gx,&gy);
        f = f-xi;
        g = g-eta;
        delta_x = (-f*gy+g*fy)/(fx*gy-fy*gx);
        delta_y = (-g*fx+f*gx)/(fx*gy-fy*gx);
        object_x = object_x+delta_x;
        object_y = object_y+delta_y;
        if ((fabs(delta_x) < tolerance) && (fabs(delta_y) < tolerance)) break;
    }
    /*
     *  Convert mm from plate center to pixels
     */
    *x = (header->ppo_coeff[2]-object_x*1000.0)/header->x_pixel_size;
    *y = (header->ppo_coeff[5]+object_y*1000.0)/header->y_pixel_size;

} /* amdinv */

/******************************************************************************
* Module:                                                                     *
*    amdpos                                                                   *
*                                                                             *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*       Routine to convert x,y to RA,Dec using the CALOBCC solution.          *
*                                                                             *
* Return Value:                                                               *
*    none                                                                     *
*                                                                             *
* Arguments:                                                                  *
*                                                                             *
*    Input Arguments:                                                         *
*       h                                                                     *
*          HEADER, Reference                                                  *
*             Structure with header information.                              *
*       x                                                                     *
*          double, Value                                                      *
*             Input x position.                                               *
*       y                                                                     *
*          double, Value                                                      *
*             Input y position.                                               *
*       mag                                                                   *
*          double, Value                                                      *
*             Magnitude for the calibration solution.                         *
*       col                                                                   *
*          double, Value                                                      *
*             Color for the calibration solution.                             *
*    Input/Output Arguments:                                                  *
*       none                                                                  *
*    Output Arguments:                                                        *
*       ra                                                                    *
*          double, Reference.                                                 *
*             The resulting right asension.                                   *
*       dec                                                                   *
*          double, Reference                                                  *
*             The resulting declination.                                      *
*                                                                             *
* History:                                                                    *
*    Created    : 21-SEP-90 by R. White,                                      *
*       Converted from IDL routine to Fortran.                                *
*    Delivered  : 31-JUL-91 by R. White,                                      *
*       Converted from Fortran to C.                                          *
*                                                                             *
*    Updated    : 31-JAN-94 by J. Doggett,                                    *
*       Adopted into software used to facilitate the accessing of data from   *
*          the CD-ROM set of the Digitized Sky Survey.                        *
*    Redelivered: 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
******************************************************************************/
extern void
amdpos(h,x,y,mag,colour,ra,dec)
HEADER h;
double x,y,mag,colour;
double *ra, *dec;
{
double ox,oy,ox2,oy2,ox3,oy3,object_xi,object_eta;

    /*
     *  Convert x,y from pixels to mm measured from plate center
     */

/****************
* amdpos page 2 *
****************/

    ox = (h->ppo_coeff[2]   - x*h->x_pixel_size)/1000.0;
    oy = (y*h->y_pixel_size - h->ppo_coeff[5]  )/1000.0;
    ox2 = ox*ox;
    oy2 = oy*oy;
    ox3 = ox*ox2;
    oy3 = oy*oy2;
    /*
     *  Compute standard coordinates from x,y and plate model
     */
    object_xi =
     h->amd_x_coeff[ 0]*ox                    + h->amd_x_coeff[ 1]*oy     +
     h->amd_x_coeff[ 2]                       + h->amd_x_coeff[ 3]*ox2    +
     h->amd_x_coeff[ 4]*ox*oy                 + h->amd_x_coeff[ 5]*oy2    +
     h->amd_x_coeff[ 6]*(ox2+oy2)             + h->amd_x_coeff[ 7]*ox3    +
     h->amd_x_coeff[ 8]*ox2*oy                + h->amd_x_coeff[ 9]*ox*oy2 +
     h->amd_x_coeff[10]*oy3                   +
     h->amd_x_coeff[11]*ox*(ox2+oy2)          +
     h->amd_x_coeff[12]*ox*(ox2+oy2)*(ox2*oy2)+
     h->amd_x_coeff[13]*mag                   + h->amd_x_coeff[14]*mag*mag+
     h->amd_x_coeff[15]*mag*mag*mag           + h->amd_x_coeff[16]*mag*ox +
     h->amd_x_coeff[17]*mag*(ox2+oy2)         +
     h->amd_x_coeff[18]*mag*ox*(ox2+oy2)      +
     h->amd_x_coeff[19]*colour;

    object_eta =
     h->amd_y_coeff[ 0]*oy                    + h->amd_y_coeff[ 1]*ox     +
     h->amd_y_coeff[ 2]                       + h->amd_y_coeff[ 3]*oy2    +
     h->amd_y_coeff[ 4]*oy*ox                 + h->amd_y_coeff[ 5]*ox2    +
     h->amd_y_coeff[ 6]*(oy2+ox2)             + h->amd_y_coeff[ 7]*oy3    +
     h->amd_y_coeff[ 8]*oy2*ox                + h->amd_y_coeff[ 9]*oy*ox2 +
     h->amd_y_coeff[10]*ox3                   +
     h->amd_y_coeff[11]*oy*(oy2+ox2)          +
     h->amd_y_coeff[12]*oy*(oy2+ox2)*(oy2*ox2)+
     h->amd_y_coeff[13]*mag                   + h->amd_y_coeff[14]*mag*mag+
     h->amd_y_coeff[15]*mag*mag*mag           + h->amd_y_coeff[16]*mag*oy +
     h->amd_y_coeff[17]*mag*(oy2+ox2)         +
     h->amd_y_coeff[18]*mag*oy*(oy2+ox2)      +
     h->amd_y_coeff[19]*colour;
    /*
     *  Convert to RA and Dec
     *  Note that ra and dec are already pointers, so we don't need
     *  to pass by address
     */
    transtdeq(h->plt_center_ra,h->plt_center_dec,
	      object_xi,object_eta,ra,dec);
}

/******************************************************************************
* Module:                                                                     *
*    pltmodel                                                                 *
*                                                                             *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*    Computes values of a plate model and its parital derivatives for use in  *
*    computing inverse astrometric solutions.                                 *
*                                                                             *
* Return Value:                                                               *
*    none                                                                     *
*                                                                             *
* Arguments:                                                                  *
*                                                                             *
*    Input Arguments:                                                         *
*       h                                                                     *
*          HEADER, Reference                                                  *
*             Structure with header information.                              *
*       x                                                                     *
*          double, Value                                                      *
*             X pixel position.                                               *
*       y                                                                     *
*          double, Value                                                      *
*             Y pixel position.                                               *
*       mag                                                                   *
*          double, Value                                                      *
*             Magnitude                                                       *
*       color                                                                 *
*          double, Value                                                      *
*             Color                                                           *
*    Input/Output Arguments:                                                  *
*       none                                                                  *
*    Output Arguments:                                                        *
*       f                                                                     *
*          double, Reference                                                  *
*             Xi standard coordinate.                                         *
*       fx                                                                    *
*          double, Reference                                                  *
*             Derivative of xi with respect to x.                             *
*       fy                                                                    *
*          double, Reference                                                  *
*             Derivative of xi with respect to y.                             *
*       g                                                                     *
*          double, Reference                                                  *
*             eta standard coordinate.                                        *
*       gx                                                                    *
*          double, Reference                                                  *
*             Derivative of eta with respect to x.                            *
*       gy                                                                    *
*          double, Reference                                                  *
*             Derivative of eta with respect to y.                            *
*                                                                             *
* History:                                                                    *
*    Created    : 07-MAY-90 by R. White,                                      *
*       Converted from GASP IDL routine to Fortran.                           *
*    Delivered  : 30-JUL-91 by R. White,                                      *
*       Converted from Fortran to C.                                          *
*                                                                             *
*    Updated    : 13-AUG-93 by J. Doggett,                                    *
*       Adopted into software used to facilitate the accessing of data from   *
*          the CD-ROM set of the Digitized Sky Survey.                        *
*******************************************************************************

*******************
* pltmodel page 2 *
*******************

*******************************************************************************
*    Redelivered: 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
*                                                                             *
*    Updated    : 01-AUG-94 by J. Doggett,                                    *
*       Corrected error in term [11] of the derivative of Y model wrt y.      *
*    Redelivered: 30-NOV-94 by J. Doggett,                                    *
*       GetImage Version 1.1.                                                 *
******************************************************************************/
extern void
pltmodel(h,x,y,mag,colour,f,fx,fy,g,gx,gy)
HEADER h;
double x,y,mag,colour;
double *f, *fx, *fy, *g, *gx, *gy;
{
double cjunk,x4,y4;

    /*
     *  X plate model
     */
    cjunk=(x*x+y*y)*(x*x+y*y);
    x4 = (x*x)*(x*x);
    y4 = (y*y)*(y*y);
    *f=h->amd_x_coeff[0]*x                 + h->amd_x_coeff[1]*y              +
       h->amd_x_coeff[2]                   + h->amd_x_coeff[3]*x*x            +
       h->amd_x_coeff[4]*x*y               + h->amd_x_coeff[5]*y*y            +
       h->amd_x_coeff[6]*(x*x+y*y)         + h->amd_x_coeff[7]*x*x*x          +
       h->amd_x_coeff[8]*x*x*y             + h->amd_x_coeff[9]*x*y*y          +
       h->amd_x_coeff[10]*y*y*y            + h->amd_x_coeff[11]*x*(x*x+y*y)   +
       h->amd_x_coeff[12]*x*cjunk          + h->amd_x_coeff[13]*mag           +
       h->amd_x_coeff[14]*mag*mag          + h->amd_x_coeff[15]*mag*mag*mag   +
       h->amd_x_coeff[16]*mag*x            + h->amd_x_coeff[17]*mag*(x*x+y*y) +
       h->amd_x_coeff[18]*mag*x*(x*x+y*y)  + h->amd_x_coeff[19]*colour;
    /*
     *  Derivative of X model wrt x
     */
    *fx=h->amd_x_coeff[0]                              +
        h->amd_x_coeff[3]*2.0*x                        +
        h->amd_x_coeff[4]*y                            +
        h->amd_x_coeff[6]*2.0*x                        +
        h->amd_x_coeff[7]*3.0*x*x                      +
        h->amd_x_coeff[8]*2.0*x*y                      +
        h->amd_x_coeff[9]*y*y                          +
        h->amd_x_coeff[11]*(3.0*x*x+y*y)               +
        h->amd_x_coeff[12]*(5.0*x4  +6.0*x*x*y*y+y4  ) +
        h->amd_x_coeff[16]*mag                         +
        h->amd_x_coeff[17]*mag*2.0*x                   +
        h->amd_x_coeff[18]*mag*(3.0*x*x+y*y);
    /*
     *  Derivative of X model wrt y
     */
    *fy=h->amd_x_coeff[1]                     +
        h->amd_x_coeff[4]*x                   +
        h->amd_x_coeff[5]*2.0*y               +
        h->amd_x_coeff[6]*2.0*y               +
        h->amd_x_coeff[8]*x*x                 +
        h->amd_x_coeff[9]*x*2.0*y             +
        h->amd_x_coeff[10]*3.0*y*y            +
        h->amd_x_coeff[11]*2.0*x*y            +
        h->amd_x_coeff[12]*4.0*x*y*(x*x+y*y)  +

/******************
* pltmodel page 3 *
******************/

        h->amd_x_coeff[17]*mag*2.0*y          +
        h->amd_x_coeff[18]*mag*2.0*x*y;
    /*
     *  Y plate model
     */
    *g=h->amd_y_coeff[0]*y                + h->amd_y_coeff[1]*x              +
       h->amd_y_coeff[2]                  + h->amd_y_coeff[3]*y*y            +
       h->amd_y_coeff[4]*y*x              + h->amd_y_coeff[5]*x*x            +
       h->amd_y_coeff[6]*(x*x+y*y)        + h->amd_y_coeff[7]*y*y*y          +
       h->amd_y_coeff[8]*y*y*x            + h->amd_y_coeff[9]*y*x*x          +
       h->amd_y_coeff[10]*x*x*x           + h->amd_y_coeff[11]*y*(x*x+y*y)   +
       h->amd_y_coeff[12]*y*cjunk         + h->amd_y_coeff[13]*mag           +
       h->amd_y_coeff[14]*mag*mag         + h->amd_y_coeff[15]*mag*mag*mag   +
       h->amd_y_coeff[16]*mag*y           + h->amd_y_coeff[17]*mag*(x*x+y*y) +
       h->amd_y_coeff[18]*mag*y*(x*x+y*y) + h->amd_y_coeff[19]*colour;
    /*
     *  Derivative of Y model wrt x
     */
    *gx=h->amd_y_coeff[1]                    +
        h->amd_y_coeff[4]*y                  +
        h->amd_y_coeff[5]*2.0*x              +
        h->amd_y_coeff[6]*2.0*x              +
        h->amd_y_coeff[8]*y*y                +
        h->amd_y_coeff[9]*y*2.0*x            +
        h->amd_y_coeff[10]*3.0*x*x           +
        h->amd_y_coeff[11]*2.0*x*y           +
        h->amd_y_coeff[12]*4.0*x*y*(x*x+y*y) +
        h->amd_y_coeff[17]*mag*2.0*x         +
        h->amd_y_coeff[18]*mag*y*2.0*x;
    /*
     *  Derivative of Y model wrt y
     */
    *gy=h->amd_y_coeff[0]                              +
        h->amd_y_coeff[3]*2.0*y                        +
        h->amd_y_coeff[4]*x                            +
        h->amd_y_coeff[6]*2.0*y                        +
        h->amd_y_coeff[7]*3.0*y*y                      +
        h->amd_y_coeff[8]*2.0*y*x                      +
        h->amd_y_coeff[9]*x*x                          +
        h->amd_y_coeff[11]*(x*x+3.0*y*y)               +
        h->amd_y_coeff[12]*(5.0*y4  +6.0*x*x*y*y+x4  ) +
        h->amd_y_coeff[16]*mag                         +
        h->amd_y_coeff[17]*mag*2.0*y                   +
        h->amd_y_coeff[18]*mag*(x*x+3.0*y*y);

} /* pltmodel */

/******************************************************************************
* Module:                                                                     *
*    ppoinv                                                                   *
*                                                                             *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*    Computes the x,y position on the plate from the right ascension and      *
*    declination using the PDS orientation solution.                          *
*                                                                             *
* Return Value:                                                               *
*    none                                                                     *
*                                                                             *
* Arguments:                                                                  *
*                                                                             *
*    Input Arguments:                                                         *
*       header                                                                *
*          HEADER, Reference                                                  *
*             Structure with header information.                              *
*       ra                                                                    *
*          double, Value                                                      *
*             The right asension of the coordinates to convert to x-y.        *
*       dec                                                                   *
*          double, Value                                                      *
*             The declination of the coordinates to convert to x-y.           *
*    Input/Output Arguments:                                                  *
*       none                                                                  *
*    Output Arguments:                                                        *
*       x                                                                     *
*          float, Reference                                                   *
*             Result x position.                                              *
*       y                                                                     *
*          float, Reference                                                   *
*             Result y position.                                              *
*                                                                             *
* History:                                                                    *
*    Created    : 07-MAY-90 by R. White,                                      *
*       Converted from GASP IDL routine to Fortran.                           *
*    Delivered  : 30-JUL-91 by R. White,                                      *
*       Converted from Fortran to C.                                          *
*                                                                             *
*    Updated    : 13-AUG-93 by J. Doggett,                                    *
*       Adopted into software used to facilitate the accessing of data from   *
*          the CD-ROM set of the  Digitized Sky Survey.                       *
*    Updated    : 14-AUG-93 by J. Doggett,                                    *
*       Modified to reflect change in calling sequence to traneqstd function. *
*    Redelivered: 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
******************************************************************************/
extern void
ppoinv(header,ra,dec,x,y)
HEADER header;
double ra, dec;
float *x,*y;
{
double object_xi, object_eta, object_x, object_y;

    /*
     *  Convert RA and Dec to standard coords.
     */
    traneqstd(header->plt_center_ra,header->plt_center_dec,ra,
                                                  dec,&object_xi,&object_eta);


/****************
* ppoinv page 2 *
****************/

    /*
     *  Convert st.coords from arcsec to radians
     */
    object_xi  = object_xi /ARCSECONDS_PER_RADIAN;
    object_eta = object_eta/ARCSECONDS_PER_RADIAN;

    /*
     *  Compute PDS coordinates from solution
     */
    object_x = header->ppo_coeff[0]*object_xi
        + header->ppo_coeff[1]*object_eta + header->ppo_coeff[2];
    object_y = header->ppo_coeff[3]*object_xi
        + header->ppo_coeff[4]*object_eta + header->ppo_coeff[5];
    /*
     * Convert x,y from microns to pixels
     */
    *x = object_x/header->x_pixel_size;
    *y = object_y/header->y_pixel_size;

} /* ppoinv */

/******************************************************************************
* Module:                                                                     *
*    xypos                                                                    *
* File:                                                                       *
*    astrmcal.c                                                               *
*                                                                             *
* Description:                                                                *
*    Computes the x,y position from right ascension and declination. The      *
*    CALOBCC solution is used if available. Otherwise, the DIG25CC            *
*    orientation solution is used.                                            *
*                                                                             *
* Return Value:                                                               *
*    none                                                                     *
*                                                                             *
* Arguments:                                                                  *
*                                                                             *
*    Input Arguments:                                                         *
*       header                                                                *
*          HEADER, Reference                                                  *
*             Structure with header information.                              *
*       ra                                                                    *
*          double, Value                                                      *
*             The right asension of the coordinates to convert to x-y.        *
*       dec                                                                   *
*          double, Value                                                      *
*             The declination of the coordinates to convert to x-y.           *
*       mag                                                                   *
*          float, Value                                                       *
*             Magnitude for the calibration solution.                         *
*       col                                                                   *
*          float, Value                                                       *
*             Color for the calibration solution.                             *
*    Input/Output Arguments:                                                  *
*       none                                                                  *
*    Output Arguments:                                                        *
*       x                                                                     *
*          float, Reference                                                   *
*             Result x position.                                              *
*       y                                                                     *
*          float, Reference                                                   *
*             Result y position.                                              *
*                                                                             *
* History:                                                                    *
*    Created    : 07-MAY-90 by R. White,                                      *
*       Converted from GASP IDL routine to Fortran.                           *
*    Delivered  : 30-JUL-91 by R. White,                                      *
*       Converted from Fortran to C.                                          *
*                                                                             *
*    Updated    : 13-AUG-93 by J. Doggett,                                    *
*       Adopted into software used to facilitate the accessing of data from   *
*          the CD-ROM set of the Digitized Sky Survey.                        *
*    Redelivered: 01-MAR-94 by J. Doggett,                                    *
*       GetImage Version 1.0                                                  *
******************************************************************************/
extern void
xypos(header,ra,dec,mag,col,x,y)
HEADER header;
double ra,dec;
float mag,col;
float *x,*y;
{
    if (header->amd_flag == 'T') {
        amdinv(header,ra,dec,mag,col,x,y);
    } else {

/***************
* xypos page 2 *
***************/

        ppoinv(header,ra,dec,x,y);
    }
} /* xypos */
